/*
 * Utilities for handling the UID range allocated for non-system user
 */

#include <stdio.h>
#include <sys/stat.h>
#ifdef __linux__
#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>
#endif

#include "kdm-minmaxuid.h"

static void internalMinMaxUids(int *minUid, int *maxUid, int staticValues)
{
    struct stat st;

#ifdef __linux__
    if (!stat("/etc/debian_version", &st)) { /* debian */
        *minUid = 1000;
        *maxUid = 29999;
    } else if (!stat("/usr/portage", &st)) { /* gentoo */
        *minUid = 1000;
        *maxUid = 65000;
    } else if (!stat("/etc/mandrake-release", &st)) { /* mandrake - check before redhat! */
        *minUid = 500;
        *maxUid = 65000;
    } else if (!stat("/etc/redhat-release", &st)) { /* redhat */
	if (!staticValues && !stat("/etc/login.defs", &st))
	    *minUid = -1;
	else
	    *minUid = 100;
        *maxUid = 65000;
    } else /* if (!stat("/etc/SuSE-release", &st)) */ { /* suse */
        *minUid = 500;
        *maxUid = 65000;
    }
#else
    *minUid = 1000;
    *maxUid = 65000;
#endif
}

void getDefaultMinMaxUids(int *minUid, int *maxUid)
{
    internalMinMaxUids(minUid, maxUid, 0);
}

#ifdef __linux__
static int getLoginDefsMinUid(void)
{
    int res;
    char *buf;
    size_t buf_size;
    ssize_t line_len;
    regex_t re;
    FILE *f;

    f = fopen("/etc/login.defs", "r");
    if (f == NULL)
	return -1;

    /* No, this is not particularly nice, but "compatible" is more important
       than "beautiful". */
    if (regcomp(&re,
		"^[ \t]*"	/* Initial whitespace */
		"([^ \t]+)"	/* Variable name */
		"[ \t][ \t\"]*"	/* Separator - yes, may have multiple "s */
		"("
		"([^\"]*)\".*"	/* Value, case 1 - terminated by " */
		/* Value, case 2 - only drop trailing [[:space:]] */
		"|([^\"]*[^[:space:]])?[[:space:]]*"
		")$", REG_EXTENDED) != 0)
	abort();

    buf = NULL;
    buf_size = 0;
    res = -1;
    while ((line_len = getline(&buf, &buf_size, f)) > 0) {
	regmatch_t match[5];
	const char *name, *value;

	if (line_len > 0 && buf[line_len - 1] == '\n')
	    buf[line_len - 1] = '\0';

	if (regexec(&re, buf, sizeof(match) / sizeof(*match), match, 0) != 0)
	    continue;

	assert(match[1].rm_so != -1);
	name = buf + match[1].rm_so;
	if (*name == '#')
	    continue;
	/* The "separator" is at least one character long, so we can safely
	   overwrite it with '\0'. */
	buf[match[1].rm_eo] = '\0';

	if (match[3].rm_so != -1) {
	    value = buf + match[3].rm_so;
	    buf[match[3].rm_eo] = '\0';
	} else if (match[4].rm_so != -1) {
	    value = buf + match[4].rm_so;
	    buf[match[4].rm_eo] = '\0';
	} else
	    value = "";

	/* fprintf(stderr, "\tMatched: %s = %s!\n", name, value); */

	if (strcmp(name, "UID_MIN") == 0) {
	    intmax_t imax;
	    char *p;

	    errno = 0;
	    imax = strtoimax(value, &p, 10);
	    if (errno == 0 && *p == '\0' && p != value && (uid_t)imax == imax)
		res = imax;
	}
    }
    free(buf);

    regfree(&re);
    fclose(f);

    return res;
}
#endif

int interpretMinUid(int minUid)
{
    int res, dummy;

    if (minUid != -1)
	return minUid;
#ifdef __linux__
    res = getLoginDefsMinUid();
    if (res != -1)
	return res;
#endif
    internalMinMaxUids(&res, &dummy, 1);
    return res;
}

#ifdef TESTING
int main(void)
{
    printf("interpretMinUid returns %d\n", interpretMinUid(-1));
    return 0;
}
#endif
